@@ -48,6 +48,8 @@ let big_int_of_value v =
   match v with
   | Int n -> n
   | Term {t_node = Tconst c } -> big_int_of_const c
+  | Term { t_node = Tapp (ls,[{ t_node = Tconst c }]) }
+    when ls_compare ls !ls_minus = 0 -> BigInt.minus (big_int_of_const c)
   | _ -> raise NotNum
 
 
@@ -430,7 +432,7 @@ let rec reduce engine c =
                 Term
                   (t_label_copy orig
                      (t_if t1 (t_subst sigma t2) (t_subst sigma t3))) :: st;
-              cont_stack = rem ;
+              cont_stack = rem;
             }
         end
       | Int _ -> assert false (* would be ill-typed *)
@@ -455,7 +457,7 @@ let rec reduce engine c =
   | [], (Knot,_) :: _ -> assert false
   | Int _ :: _ , (Knot,_) :: _ -> assert false
   | (Term t) :: st, (Knot, orig) :: rem ->
-    { value_stack = Term (t_label_copy orig (t_not t)) :: st;
+    { value_stack = Term (t_label_copy orig (t_not_simp t)) :: st;
       cont_stack = rem;
     }
   | st, (Kapp(ls,ty), orig) :: rem ->
@@ -512,8 +514,13 @@ and reduce_match st u ~orig tbl sigma cont =
       with NoMatch -> iter rem
   in
   try iter tbl with Undetermined ->
+    let dmy = t_var (create_vsymbol (Ident.id_fresh "__dmy") (t_type u)) in
+    let tbls = match t_subst sigma (t_case dmy tbl) with
+      | { t_node = Tcase (_,tbls) } -> tbls
+      | _ -> assert false
+    in
     { value_stack =
-        Term (t_label_copy orig (t_subst sigma (t_case u tbl))) :: st;
+        Term (t_label_copy orig (t_case u tbls)) :: st;
       cont_stack = cont;
     }
 
@@ -700,9 +707,9 @@ and reduce_app_no_equ engine st ls ~orig ty rem_cont =
     }
   with Not_found | Undetermined ->
     let args = List.map term_of_value args in
-    try
-      let d = Ident.Mid.find ls.ls_name engine.known_map in
-      let rewrite () =
+    let d = try Ident.Mid.find ls.ls_name engine.known_map
+      with Not_found -> assert false in
+    let rewrite () =
       (* try a rewrite rule *)
         begin
           try
@@ -747,7 +754,9 @@ and reduce_app_no_equ engine st ls ~orig ty rem_cont =
               cont_stack = (Keval(rhs,mv),orig) :: rem_cont;
             }
           with Irreducible ->
-            raise Not_found
+            { value_stack =
+                Term (t_label_copy orig (t_app ls args ty)) :: rem_st;
+              cont_stack = rem_cont; }
         end in
       match d.Decl.d_node with
       | Decl.Dtype _ | Decl.Dprop _ -> assert false
@@ -772,13 +781,13 @@ and reduce_app_no_equ engine st ls ~orig ty rem_cont =
         rewrite ()
       | Decl.Ddata dl ->
         (* constructor or projection *)
-        match args with
+        begin try match args with
         | [ { t_node = Tapp(ls1,tl1) } ] ->
           (* if ls is a projection and ls1 is a constructor,
              we should compute that projection *)
           let rec iter dl =
             match dl with
-            | [] -> raise Not_found
+            | [] -> raise Exit
             | (_,csl) :: rem ->
               let rec iter2 csl =
                 match csl with
@@ -800,16 +809,13 @@ and reduce_app_no_equ engine st ls ~orig ty rem_cont =
                           iter3 prs tl1
                       | None::prs, _::tl1 ->
                         iter3 prs tl1
-                      | _ -> raise Not_found
+                      | _ -> raise Exit
                     in iter3 prs tl1
                   else iter2 rem2
               in iter2 csl
           in iter dl
-        | _ -> raise Not_found
-    with Not_found ->
-      { value_stack = Term (t_label_copy orig (t_app ls args ty)) :: rem_st;
-        cont_stack = rem_cont;
-      }
+        | _ -> raise Exit
+        with Exit -> rewrite () end
 
 
 and reduce_equ (* engine *) ~orig st v1 v2 cont =
@@ -987,8 +993,19 @@ let extract_rule _km t =
       | Decl.Dparam _ | Decl.Dind _ -> ()
   in
 *)
-  (* TODO : verifier que les variables de droite, aussi bien term que type,
-     apparaissent a gauche *)
+
+  let check_vars acc t1 t2 =
+    (* check that quantified variables all appear in the lefthand side *)
+    let vars_lhs = t_vars t1 in
+    if Svs.exists (fun vs -> not (Mvs.mem vs vars_lhs)) acc
+    then raise (NotARewriteRule "lhs should contain all variables");
+    (* check the same with type variables *)
+    if not
+         (Ty.Stv.subset
+            (t_ty_freevars Ty.Stv.empty t2) (t_ty_freevars Ty.Stv.empty t2))
+    then raise (NotARewriteRule "lhs should contain all type variables")
+
+  in
 
   let rec aux acc t =
     match t.t_node with
@@ -998,14 +1015,20 @@ let extract_rule _km t =
       | Tbinop(Tiff,t1,t2) ->
         begin
           match t1.t_node with
-            | Tapp(ls,args) -> (* check_ls ls; *) acc,ls,args,t2
+            | Tapp(ls,args) ->
+               (* check_ls ls; *)
+               check_vars acc t1 t2;
+               acc,ls,args,t2
             | _ -> raise
               (NotARewriteRule "lhs of <-> should be a predicate symbol")
         end
       | Tapp(ls,[t1;t2]) when ls == ps_equ ->
         begin
           match t1.t_node with
-            | Tapp(ls,args) -> (* check_ls ls; *) acc,ls,args,t2
+            | Tapp(ls,args) ->
+               (* check_ls ls; *)
+               check_vars acc t1 t2;
+               acc,ls,args,t2
             | _ -> raise
               (NotARewriteRule "lhs of = should be a function symbol")
         end
